@@ -1,24 +1,54 @@
-#' Function to handle animal track data, organized as \code{"trip"}s
+#' Function to handle animal track data, organized as \code{trip} objects
 #'
 #'
-#' Create an object of class \code{"trip"}, extending the basic functionality
+#' Create an object of class \code{trip}, extending the basic functionality
 #' of \code{\link[sp]{SpatialPointsDataFrame}} by specifying the data columns
 #' that define the "TimeOrdered" quality of the records.
 #'
-#'
+#' The original form of `trip()` required very strict input as a 'SpatialPointsDataFrame' and 
+#' specifying which were the time and ID columns, but the input can be more flexible. If the object is a 
+#' grouped data frame ('dplyr-style') then the (first) grouping is assumed to define individual trips and that 
+#' columns 1, 2, 3 are the x-, y-, time-coordinates in that order. It can also be a \code{trip} object for
+#' redefining \code{TORnames}.  
+#' 
+#' The [trip()] function can ingest `track_xyt`, `telemetry`, `SpatialPointsDataFrame`, `sf`, 
+#' `trackeRdata`, `grouped_df`, `data.frame`, `tbl_df`, `mousetrap`, and in some cases
+#' lists of those objects. Please get in touch if you think something that should work does not. 
+#'  
+#' Track data often contains problems, with missing values in location or time, 
+#' times out of order or with duplicated times. The `correct_all` argument is 
+#' set to `TRUE` by default and will report any inconsistencies. Data really should
+#' be checked first rather than relying on this auto-cleanup. The following problems are common: 
+#' * duplicated records (every column with the same value in another row)
+#' * duplicated date-time values
+#' * missing date-time values, or missing x or y coordinates
+#' * records out of order within trip ID
+#' 
+#' For some data types there's no formal structure, but a simple convention such as
+#' a set of names in a data frame. For example, the VTrack package has `AATAMS1` which may be
+#' turned into a trip with 
+#' `trip(AATAMS1 %>% dplyr::select(longitude, latitude, timestamp, tag.ID, everything())`
+#' In time we can add support for all kinds of variants, detected by the names and contents. 
+#' 
+#' 
+#' See [Chapter 2 of the trip thesis](https://eprints.utas.edu.au/12273/) for more details. 
 #' @name trip-methods
 #' @aliases trip-methods trip trip,SpatialPointsDataFrame,ANY-method
+#' trip,SpatialPointsDataFrame,TimeOrderedRecords-method
 #' trip,ANY,TimeOrderedRecords-method trip,trip,ANY-method
-#' trip,trip,TimeOrderedRecords-method [,trip-method [,trip,ANY,ANY,ANY-method 
+#' trip,grouped_df,ANY-method trip,data.frame,ANY-method trip,track_xyt,ANY-method
+#' trip,trackeRdata,ANY-method trip,mousetrap,ANY-method trip,sf,ANY-method
+#' trip,telemetry,ANY-method trip,list,ANY-method
+#' trip,trip,TimeOrderedRecords-method split,trip,ANY-method [,trip-method [,trip,ANY,ANY,ANY-method 
 #' [[<-,trip,ANY,missing-method trip<-,data.frame,character-method
-#' @param obj A \code{\link[sp]{SpatialPointsDataFrame}}, or an object that can
-#' be coerced to one, containing at least two columns with the DateTime and ID
-#' data as per \code{TORnames}.  It can also be a \code{trip} object for
-#' redefining \code{TORnames}.
+#' @param obj A data frame, a grouped data frame or a \code{\link[sp]{SpatialPointsDataFrame}}
+#' containing at least two columns with the DateTime and ID data as per \code{TORnames}.  See 
+#' Details. 
 #' @param TORnames Either a \code{TimeOrderedRecords} object, or a 2-element
 #' character vector specifying the DateTime and ID column of \code{obj}
 #' @param value A 4-element character vector specifying the X, Y, DateTime coordinates 
 #' and ID of \code{obj}. 
+#' @param correct_all logical value, if `TRUE` the input data is corrected for common problems
 #' @return
 #'
 #' A trip object, with the usual slots of a
@@ -35,7 +65,8 @@
 #'
 #' \item{trip}{\code{signature(obj="SpatialPointsDataFrame",
 #' TORnames="ANY")}}The main construction.
-#'
+#' \item{trip}{\code{signature(obj="SpatialPointsDataFrame",
+#' TORnames="TimeOrderedRecords")}} Object and TimeOrdered records class
 #' \item{trip}{\code{signature(obj="ANY", TORnames="TimeOrderedRecords")}:
 #' create a \code{trip} object from a data frame.}
 #'
@@ -55,12 +86,39 @@
 #'
 #'
 #' d <- data.frame(x=1:10, y=rnorm(10), tms=Sys.time() + 1:10, id=gl(2, 5))
-#' coordinates(d) <- ~x+y
+#' 
+#' ## the simplest way to create a trip is by order of columns
+#' 
+#' trip(d)
+#' 
+#' ## or a grouped data frame can be used, the grouping is used as the trip ID
+#' ## library(dplyr)
+#' ## # use everything() to keep all other columns
+#' ## d %>% group_by(id) %>% select(x, y, tms, everything())
+#' 
+#' sp::coordinates(d) <- ~x+y
 #' ## this avoids complaints later, but these are not real track data (!)
-#' proj4string(d) <- CRS("+proj=laea +ellps=sphere")
+#' sp::proj4string(d) <- sp::CRS("+proj=laea +ellps=sphere")
 #' (tr <- trip(d, c("tms", "id")))
 #'
-#' ## don't want adehabitatMA to be loaded as a requirement here
+#'  ## real world data in CSV
+#' mi_dat <- read.csv(system.file("extdata/MI_albatross_sub10.csv", package = "trip"), 
+#'             stringsAsFactors = FALSE)
+#' ## installed subset because the data is quite dense
+#' ## mi_dat <- mi_dat[seq(1, nrow(mi_dat), by = 10), ]
+#' mi_dat$gmt <- as.POSIXct(mi_dat$gmt, tz = "UTC")
+#' mi_dat$sp_id <-  sprintf("%s%s_%s_%s", mi_dat$species, 
+#'          substr(mi_dat$breeding_status, 1, 1), mi_dat$band, mi_dat$tag_ID)
+#' sp::coordinates(mi_dat) <- c("lon", "lat")
+#' ## there are many warnings, but the outcome is fine 
+#' ## (sp_id == 'WAi_14030938_2123' has < 3 locations as does LMi_12143650_14257)
+#' mi_dat <- trip(mi_dat, c("gmt", "sp_id") )
+#' plot(mi_dat, pch = ".")
+#' #lines(mi_dat)  ## ugly
+#' 
+#' mi_dat_polar <- sp::spTransform(mi_dat, "+proj=stere +lat_0=-90 +lon_0=154 +datum=WGS84")
+#' plot(mi_dat_polar, pch = ".") 
+#' lines(mi_dat_polar)
 #' \dontrun{
 #' ## a simple example with the common fixes required for basic track data
 #'
@@ -108,17 +166,17 @@
 #'    library(raster)
 #'
 #'    ## 3 degrees either side (for half a zone . . .)
-#'    ext <- as(extent(spTransform(porpoise, CRS(proj4string(wrld_simpl)))) + 3, "SpatialPolygons")
+#'    ext <- as(extent(sp::spTransform(porpoise, CRS(proj4string(wrld_simpl)))) + 3, "SpatialPolygons")
 #'    proj4string(ext) <- CRS(proj4string(wrld_simpl))
 #'    ## crop to the buffered tracks, and project to its native CRS
-#'    w <- spTransform(gIntersection(wrld_simpl[grep("United States", wrld_simpl$NAME), ], ext),
+#'    w <- sp::spTransform(gIntersection(wrld_simpl[grep("United States", wrld_simpl$NAME), ], ext),
 #'     CRS(proj4string(porpoise)))
 #'
 #'    plot(w)
 #'    lines(porpoise)
 #' }
 setGeneric("trip",
-             function(obj, TORnames) standardGeneric("trip"))
+             function(obj, TORnames, correct_all = TRUE) standardGeneric("trip"))
 
 if (!isGeneric("points"))
   setGeneric("points",
@@ -195,30 +253,204 @@ getTORnames <- function(obj) obj@TOR.columns
 ##' @rdname trip-accessors
 ##' @export
 getTimeID <- function(obj) as.data.frame(obj)[, getTORnames(obj)]
+assume_if_longlat <- function(x) {
+  if (is.na(x@proj4string@projargs) && raster::couldBeLonLat(x, warnings = FALSE)) {
+    warning("input looks like longitude/latitude data, assuming +proj=longlat +datum=WGS84")
+    x@proj4string@projargs <- "+proj=longlat +ellps=WGS84 +datum=WGS84 +no_defs +towgs84=0,0,0"
+  }
+  x
+}
+trip.grouped_df <- function(obj, ..., crs = NULL) {
+  group_var <- setdiff(names(attr(obj, "groups")), ".rows")
+  if (length(group_var) > 1) {
+    group_var <- group_var[1]
+    warning(sprintf("data is grouped by more than one variable, assuming '%s' as the correct one", group_var))
+  }
+  tor <- c(names(obj)[3], group_var)
+  if (!inherits(obj[[tor[1]]], "POSIXct")) stop(sprintf("3rd column [%s] must be date-time", tor[1]))
+  obj <- as.data.frame(as.list(obj), stringsAsFactors = FALSE) ## remove grouping
+  sp::coordinates(obj) <- names(obj)[1:2]
+  if (!is.null(crs)) sp::proj4string(obj)
+  trip(obj, tor, ...)
+}
 
+setMethod("trip", signature(obj = "list", TORnames = "ANY"), 
+          function(obj, TORnames, correct_all = TRUE) {
+            ## a dirty trick but will work for some stuffs
+            chk <- try(trip(obj[[1]]), silent = TRUE)
+            ## this a bit slow because trips get created, need rbind for trip
+            if (!inherits(chk, "try-error")) {
+              out <- do.call(rbind, lapply(obj, trip))
+              tor <- getTORnames(chk)
+              
+            } else {
+              print("problem with list of this type")
+              stop(chk) ##sprintf("cannot interpret (list of) type %s", paste(class(obj[[1]]), collapse = ", ")))
+            }
+            trip(out, tor)
+          })
+setMethod("trip", signature(obj = "telemetry", TORnames = "ANY"), 
+          function(obj, TORnames, correct_all = TRUE) {
+            telemetry2trip(obj)
+          })
+setMethod("trip", signature(obj="sf", TORnames="ANY"),
+          function(obj, TORnames, correct_all = TRUE) {
+            if (missing(TORnames)) TORnames <- names(obj)[1:2]
+            gcol <- attr(obj, "sf_column")
+            cls <- class(obj[[gcol]])[1]
+            stopifnot(cls %in% c("sfc_POINT", "sfc_MULTIPOINT"))
+            xy <- do.call(rbind, unclass(obj[[gcol]]))
+            p4 <- attr(obj[[gcol]], "crs")$proj4string
+            idx <- NULL
+            if (cls == "sfc_MULTIPOINT") {
+              stop("MULTIPOINT not yet supported")  ## unclear what to do, unless tor[1] is the *offset* for XYZ[,3]?
+              ni <- unlist(lapply(obj[[gcol]], function(a) dim(a)[1]))
+              idx <- rep(seq_len(nrow(obj)), ni)
+            } 
+            obj[[gcol]] <- NULL
+            obj <- as.data.frame(as.list(unclass(obj)), stringsAsFactors = FALSE)
+            if (!is.null(idx)) obj <- obj[idx, ]
+            if (correct_all) {
+              
+              obj <- force_internal(obj, TORnames)
+            }
+            coordnames <- utils::tail(make.names(c(names(obj), c("X", "Y")), unique = TRUE), 2)
+            obj[[coordnames[1L]]] <- xy[,1L]
+            obj[[coordnames[2L]]] <- xy[,2L]
+            
+            sp::coordinates(obj) <- coordnames
+            sp::proj4string(obj) <- sp::CRS(p4)
+            out <- new("trip", obj, TimeOrderedRecords(TORnames))
+            assume_if_longlat(out)
+          })
 
+setMethod("trip", signature(obj = "mousetrap"), 
+          function(obj, TORnames, correct_all = TRUE) {
+            dat <- data.frame(xpos = as.vector(t(obj$trajectories[,,2L])), 
+                              ypos = as.vector(t(obj$trajectories[,,3L])), 
+                        timestamps = as.vector(t(obj$trajectories[,,1L])))
+            dat$timestamps <- ISOdatetime(1970, 1, 1, 0, 0, 0, tz = "UTC") + dat$timestamps
+            idx <- rep(seq_len(nrow(obj$data)), ncol(obj$trajectories))
+            dat$id <- rownames(obj$data)[idx] 
+            warning("assuming UNIX epoch for timestamp, where zero is 1970-01-01 00:00:00 UTC")
+            dat <- cbind(dat, obj$data[idx, ])
+            bad <- is.na(dat$xpos) | is.na(dat$ypos) | is.na(dat$timestamps) | is.na(dat$id)
+            if (sum(bad) > 0) {
+              warning(sprintf("removing %i records with missing coordinate values", sum(bad)))
+              dat <- dat[!bad, ]
+            }
+            sp::coordinates(dat) <- c("xpos", "ypos")
+            trip(dat, c("timestamps", "id"))
+          })
+setMethod("trip", signature(obj = "trackeRdata"), 
+          function(obj, TORnames, correct_all = TRUE) {
+            ns <- unlist(lapply(obj, function(df) dim(df)[1]))
+            
+            time <- do.call(c, lapply(obj, function(a) attr(a, "index")))
+            d <- data.frame(sport = rep(attr(obj, "sport"),  ns), 
+                            utc = time, run_id = rep(seq_along(ns), ns), stringsAsFactors = FALSE)
+            dat <- cbind(d, do.call(rbind, lapply(obj, unclass)))
+            ## remove any NA coords ...
+            bad <- (is.na(dat$longitude) | is.na(dat$longitude) |  is.na(dat$utc))
+            if (sum(bad) > 0) {
+              warning(sprintf("removing %i records with missing coordinate values", sum(bad)))
+              dat <- dat[!bad, ]
+            }
+            sp::coordinates(dat) <- c("longitude", "latitude")
+            sp::proj4string(dat) <- sp::CRS(.llproj())
+            
+            trip(dat, c("utc", "run_id"))
+          })
+setMethod("trip", signature(obj="track_xyt", TORnames= "ANY"),
+          function(obj, TORnames, correct_all = TRUE) {
+            if (missing(TORnames)) TORnames <- c("t_", "id")
+            TOR <- TimeOrderedRecords(TORnames)
+            proj <- attr(obj, "crs")
+            obj <- as.data.frame(as.list(obj), stringsAsFactors = FALSE)
+            sp::coordinates(obj) <- c("x_", "y_")
+            sp::proj4string(obj) <- proj
+            trip(obj, TOR, correct_all = correct_all)
+          })
+setMethod("trip", signature(obj="grouped_df", TORnames= "ANY"),
+          function(obj, TORnames, correct_all = TRUE) {
+            trip.grouped_df(obj, correct_all = correct_all)
+          })
+setMethod("trip", signature(obj="data.frame",  TORnames= "ANY"),
+          function(obj, TORnames, correct_all = TRUE) {
+            ## asumme input is x, y, time, ID
+            
+            tor <- names(obj)[3:4]
+            if (is.factor(obj[[tor[1]]])) {
+              obj[[tor[1]]] <- levels(obj[[tor[1]]])[obj[[tor[1]]]]
+            }
+            if (is.character(obj[[tor[1]]])) {
+              obj[[tor[1]]] <- as.POSIXct(obj[[tor[1]]], tz = "UTC") 
+            }
+            TORnames <- TimeOrderedRecords(tor)
+            
+            if (!is.numeric(obj[[1]]) || !is.numeric(obj[[2]])) stop("first two columns must be numeric, x,y or longitude,latitude")
+            sp::coordinates(obj) <- 1:2
+            
+            if (correct_all) {
+              
+              obj <- force_internal(obj, TORnames@TOR.columns)
+            }
+            
+            out <- new("trip", obj, TORnames)
+            assume_if_longlat(out)
+          })
+setMethod("trip", signature(obj="SpatialPointsDataFrame", TORnames="TimeOrderedRecords"),
+          function(obj, TORnames, correct_all = TRUE) {
+  
+            if (correct_all) {
+              
+              obj <- force_internal(obj, TORnames@TOR.columns)
+            }
+            
+            out <- new("trip", obj, TORnames)
+            assume_if_longlat(out)
+          })
 setMethod("trip", signature(obj="SpatialPointsDataFrame", TORnames="ANY"),
-          function(obj, TORnames) {
+          function(obj, TORnames, correct_all = TRUE) {
               if (is.factor(obj[[TORnames[2]]]))
                   obj[[TORnames[2]]] <- factor(obj[[TORnames[2]]])
-              new("trip", obj, TimeOrderedRecords(TORnames))
+              if (correct_all) {
+                #print(bbox(obj))
+                obj <- force_internal(obj, TORnames)
+              }
+              #print(bbox(obj))
+              out <- new("trip", obj, TimeOrderedRecords(TORnames))
+              #print(bbox(out))
+              assume_if_longlat(out)
           })
 
 setMethod("trip", signature(obj="ANY", TORnames="TimeOrderedRecords"),
-          function(obj, TORnames) {
-              new("trip", obj, TORnames)
+          function(obj, TORnames, correct_all = TRUE) {
+            if (correct_all) {
+              obj <- force_internal(obj, TORnames@TOR.columns)
+            }
+              out <- new("trip", obj, TORnames)
+              assume_if_longlat(out)
           })
 
 setMethod("trip", signature(obj="trip", TORnames="TimeOrderedRecords"),
-          function(obj, TORnames) {
-              new("trip",
+          function(obj, TORnames, correct_all = TRUE) {
+            if (correct_all) {
+              obj <- force_internal(obj, TORnames@TOR.columns)
+            }
+              out <- new("trip",
                   as(obj, "SpatialPointsDataFrame"),
                   TORnames)
+              assume_if_longlat(out)
           })
 
 setMethod("trip", signature(obj="trip", TORnames="ANY"),
-          function(obj, TORnames) {
-              trip(as(obj, "SpatialPointsDataFrame"), TORnames)
+          function(obj, TORnames, correct_all = TRUE) {
+            if (correct_all) {
+              obj <- force_internal(obj, TORnames)
+            }
+              out <- trip(as(obj, "SpatialPointsDataFrame"), TORnames)
+              assume_if_longlat(out)
           })
 
 triprepmethod <-   function(obj, value) {
@@ -269,8 +501,15 @@ setMethod("points", signature(x="trip"),
 setMethod("text", signature(x="trip"),
           function(x, ...) text(as(x, "SpatialPointsDataFrame"), ...))
 
-#setMethod("split", "SpatialPointsDataFrame", split.data.frame)
 
+split.trip <-  function(x, f, drop = FALSE, ...) {
+  lapply(split(x = seq_len(nrow(x)), f = f, drop = drop, ...), 
+         function(ind) x[ind, , drop = FALSE])
+}
+  
+setMethod("split", signature(x = "trip", f = "ANY"), 
+         split.trip
+          )
 ## setMethod("spTransform", signature=signature(x="trip", CRSobj="CRS"),
 ##           function(x, CRSobj, ...) tripTransform(x, CRSobj, ...))
 
@@ -305,7 +544,7 @@ setMethod("subset", signature(x="trip"),
               if (any(is.na(match(tor, names(spdf))))) {
                   msg <- paste("trip-defining Date or ID columns dropped,",
                                "reverting to SpatialPointsDataFrame\n\n")
-                  cat(msg)
+                  warning(msg)
                   return(spdf)
               } else {
                   tst <- any(tapply(spdf[[tor[1]]],
@@ -313,7 +552,7 @@ setMethod("subset", signature(x="trip"),
                   if (tst) {
                       msg <- paste("subset loses too many locations,",
                                "reverting to SpatialPointsDataFrame\n\n")
-                      cat(msg)
+                      warning(msg)
                       return(spdf)
                   } else return(trip(spdf, tor))
               }
@@ -343,6 +582,7 @@ setMethod("[", signature(x="trip"),
               if (any(is.na(i)))
                   stop("NAs not permitted in row index")
               spdf <- as(x, "SpatialPointsDataFrame")[i, j, ..., drop=drop]
+          
               tor <- getTORnames(x)
               if (is.factor(spdf[[tor[2]]]))
                   spdf[[tor[2]]] <- factor(spdf[[tor[2]]])
@@ -360,7 +600,7 @@ setMethod("[", signature(x="trip"),
                       cat(msg)
                       return(spdf)
                   } else {
-                      return(trip(spdf, tor))
+                      return(trip(spdf, tor, correct_all = F))
                   }
               }
           })
@@ -378,7 +618,7 @@ setMethod("summary", signature(object="trip"),
               ids <- tids[, 2]
               ## list of distances only, km/hr or units of projection
               dists <- .distances(object)
-              rmsspeed <- split(speedfilter(object, max.speed = 1, test = TRUE)$rms, ids)
+              #rmsspeed <- split(speedfilter(object, max.speed = 1, test = TRUE)$rms, ids)
 
               ## list of time diferences only, in hours
               dtimes <- lapply(split(time, ids), function(x) diff(unclass(x)/3600))
@@ -404,8 +644,8 @@ setMethod("summary", signature(object="trip"),
                   tripDistance <- sapply(dists, sum)
                   meanSpeed <- sapply(speeds, mean)
                   maxSpeed <- sapply(speeds, max)
-                  meanRMSspeed <- sapply(rmsspeed, mean, na.rm = TRUE)
-                  maxRMSspeed <- sapply(rmsspeed, max, na.rm = TRUE)
+                 # meanRMSspeed <- sapply(rmsspeed, mean, na.rm = TRUE)
+                 #  maxRMSspeed <- sapply(rmsspeed, max, na.rm = TRUE)
               })
               class(obj) <- "summary.TORdata"
               ## invisible(obj)
@@ -420,9 +660,9 @@ as.data.frame.summary.TORdata <- function(x, row.names = NULL, optional = FALSE,
                         tripDuration=x$tripDuration,
                         tripDistance=x$tripDistance,
                         meanSpeed = x$meanSpeed,
-                        maxSpeed = x$maxSpeed,
-                        meanRMSspeed = x$meanRMSspeed,
-                        maxRMSspeed = x$maxRMSspeed)
+                        maxSpeed = x$maxSpeed, stringsAsFactors = FALSE)
+                        #meanRMSspeed = x$meanRMSspeed,
+                        #maxRMSspeed = x$maxRMSspeed)
   dsumm
 }
 
@@ -532,7 +772,9 @@ setMethod("recenter", signature(obj="trip"),
           })
 
 
-
+#' @importFrom sp spTransform
+setMethod("spTransform", signature=signature(x="trip", CRSobj="character"),
+         function(x, CRSobj, ...) spTransform(x, sp::CRS(CRSobj), ...))
 
 setMethod("spTransform", signature("trip", "CRS"),
           function(x, CRSobj, ...) {
@@ -547,6 +789,7 @@ setMethod("spTransform", signature("trip", "CRS"),
                 stop(msg)
               }
             }
+  
             pts <- spTransform(as(x, "SpatialPointsDataFrame"),
                                CRSobj, ...)
             trip(pts, getTORnames(x))
